{{
'   fsrw 2.6 Copyright 2009  Tomas Rokicki and Jonathan Dummer
'
'   See end of file for terms of use.
'
'   This object provides FAT16/32 file read/write access on a block device.
'   Only one file open at a time.  Open modes are 'r' (read), 'a' (append),
'   'w' (write), and 'd' (delete).  Only the root directory is supported.
'   No long filenames are supported.  We also support traversing the
'   root directory.
'
'   In general, negative return values are errors; positive return
'   values are success.  Other than -1 on popen when the file does not
'   exist, all negative return values will be "aborted" rather than
'   returned.
'
'   Changes:
'       v1.1  28 December 2006  Fixed offset for ctime
'       v1.2  29 December 2006  Made default block driver be fast one
'       v1.3  6 January 2007    Added some docs, and a faster asm
'       v1.4  4 February 2007   Rearranged vars to save memory;
'                               eliminated need for adjacent pins;
'                               reduced idle current consumption; added
'                               sample code with abort code data
'       v1.5  7 April 2007      Fixed problem when directory is larger
'                               than a cluster.
'       v1.6  23 September 2008 Fixed a bug found when mixing pputc
'                               with pwrite.  Also made the assembly
'                               routines a bit more cautious.
'       v2.1  12 July 2009      FAT32, SDHC, multiblock, bug fixes
'       v2.4  26 September 2009 Added seek support.  Added clustersize.
'       v2.4a  6 October 2009   modified setdate to explicitly set year/month/etc.
'       v2.5  13 November 2009  fixed a bug on releasing the pins, added a "release" pass through function
'       v2.6  11 December 2009: faster transfer hub <=> cog, safe_spi.spin uses 1/2 speed reads, is default
'       v2.6a  3 June 2010      Modified to make compatible with FemtoBasic, uses sdspiFemto driver
'       v2.6b 24 July 2010      Added support for JEDEC flash memory, SRAM, and ADC
''            27 October 2010   Remi Veilleux: Code cleanup: delete every unneeded item and try to use less memory
                                Also merged 'Flash/SRAM WinbondDriver' inside this to use only one cog for all SPI

' Usage: include the object, and allocate a variable, then start it. Here's an example:
OBJ
  fsrw: "REM_drv_db_c3_fsrwfemto_020.spin"
  
VAR
  long ioControl[5]   ' SPI
  
PUB start
  fsrw.start(@ioControl)                                  ' Start SPI driver

}}
'
'   Constants describing FAT volumes.
'
con
   SECTORSIZE = 512
   SECTORSHIFT = 9
   DIRSIZE = 32
   DIRSHIFT = 5

CON
'' Command code and error information for I2C driver
'' (For convenience in using just OS_loaderInit, these are included here.  The "master"
''  copies are considered the ones in OS_loader and these must be kept up-to-date).

  ioReadCmd   = %00000001                      ' Read from EEPROM to HUB RAM (16 bit addresses)
  ioWriteCmd  = %00000010                      ' Write to EEPROM from HUB RAM (16 bit addresses)
  ioRead1Cmd  = %00000011                      ' Read from a device with only 8-bit addresses
  ioWrite1Cmd = %00000100                      ' Write to a device with only 8-bit addresses
  ioRead0Cmd  = %00000101                      ' Read from a device without address bytes
  ioWrite0Cmd = %00000110                      ' Write to a device without address bytes
  ioBootCmd   = %00001000                      ' Read from EEPROM to HUB RAM, then start a
                                               '  new SPIN interpreter in the COG whose ID is
                                               '  supplied in the lower 3 bits of this command
                                               '  This COG is stopped before the read is done
                                               '  unless it's the one used to execute the loader
  ioSpiInit   = %00010000                      ' Initialize the specified SPI bus and SD card
  ioSpiStop   = %00010001                      ' Change all SD card pins to inputs
  ioSpiRead   = %00010010                      ' Read one or more bytes from the SD card
  ioSpiWrite  = %00010011                      ' Write one or more bytes from the SD card
  ioSpiBoot   = %00011000                      ' Like ioBootCmd, but uses ioSpiRead for loading
  ioCmdMask   = %00011111                      ' Used to mask off command bits
  ioSpiMask   = %00010000                      ' Used to test for SPI command codes

' Options for commands

  ioNoStore   = %00100000                      ' If set, data is not stored into main memory
                                               '  If ioBootCmd or ioSpiBoot, no cogs are
                                               '   stopped and a new cog is not started.
  ioLowSpeed  = %01000000                      ' If set, I2C runs at 100KHz rather than 400KHz
  ioStopLdr   = %10000000                      ' If set, the loader's cog is stopped after a boot

' Return status

  ioWriteErr  = %10000000                      ' An error occurred during an I2C write (NAK)

  ioTestRdy   = ioCmdMask << 24                ' Used to test 1st control long for ready
  ioTestErr   = ioWriteErr << 24               ' Used to test 1st control long for write error

'' Other constants from OS_loader

  i2cBootSCL  = 28                             ' Boot EEPROM SCL pin
  bootAddr    = i2cBootSCL << 18               ' Address of boot EEPROM
  clkfreqVal  = $0000                          ' Current CLKFREQ value stored here
  clksetVal   = $0004                          ' Current CLKSET value stored here
  chksumVal   = $0005                          ' Checksum over memory stored here
  vbase       = $0008                          ' Length of Spin program loaded (# longs * 4)
  dbase       = $000A                          ' Address of start of stack (marker below)

  ' Merged from Winbond_driver:
    WrStat   = $01_00                  ' Write Status / Data
    WrData   = $02_000000              ' Write Data / Address / Data
    WrSRAM   = $02_0000                ' Write SRAM / Address / Data
    RdData   = $03_000000              ' Read Data / Address / (Data)
    RdSRAM   = $03_0000                ' Read SRAM / Address / (Data)
    Status   = $05                     ' Read Status / (Data)
    WrtEna   = $06                     ' Write Enable
    SecEra   = $20_000000              ' Erase Sector / Address
    ChipEra  = $60
    JEDEC    = $9F                     ' Return JEDEC device info
    JEDEC1M  = $01451F                 '  1MB Flash Memory (AT26DF081)
    JEDEC2M  = $1530EF                 '  2MB Flash Memory (W25X16/16A)
    JEDEC4M  = $1630EF                 '  4MB Flash Memory (W25X32)
    JEDEC8M  = $1730EF                 '  8MB Flash Memory (W25X64)

    invDev   = 0                       ' Invalid device number
    SRAM0    = 1                       ' First 32K256 SRAM
    SRAM1    = 2                       ' Second 32K256 SRAM
    FLASH    = 3                       ' Atmel or Winbond Flash
    ADC      = 4                       ' MCP3202 2 Channel 12 Bit ADC
    microSD  = 5                       ' Micro SD card
    SPI6     = 6                       ' Unspecified SPI select on header
    SPI7     = 7                       ' Unspecified SPI select on header


    Busy     = $01                     ' Busy Status

  ' SPI communication on C3
  spiCnt    =  8                      ' SPI select count 0->1->0
  spiDI     =  9                      ' SPI MOSI (or DI)
  spiDO     = 10' 7 'Replace with 10 on normal C3                     ' SPI MISO (or DO)
  spiClk    = 11                      ' SPI Clock 0->1->0
  spiSel    =  5                      ' SPI select for microSD card
  spiClr    = 25                      ' SPI select clear 1->0->1

'
'   The object that provides the block-level access.
'
var
'
'   Variables concerning the open file.
'
   long fclust ' the current cluster number
   long filesize ' the total current size of the file
   long floc ' the seek position of the file
   long frem ' how many bytes remain in this cluster from this file
   long bufat ' where in the buffer our current character is
   long bufend ' the last valid character (read) or free position (write)
   long direntry ' the byte address of the directory entry (if open for write)
   long writelink ' the byte offset of the disk location to store a new cluster
   long fatptr ' the byte address of the most recently written fat entry
   long firstcluster ' the first cluster of this file
   long filesystem
   long rootdir ' the byte address of the start of the root directory
   long rootdirend ' the byte immediately following the root directory.
   long dataregion ' the start of the data region, offset by two sectors
   long clustershift ' log base 2 of blocks per cluster
   long clustersize ' total size of cluster in bytes
   long fat1 ' the block address of the fat1 space
   long totclusters ' how many clusters in the volume
   long sectorsperfat ' how many sectors per fat
   long endofchain ' end of chain marker (with a 0 at the end)
   long pdate ' current date
   long lastread ' the block address of the buf2 contents
   long dirty ' nonzero if buf2 is dirty
   'long buf2_mem[SECTORSIZE>>2]  ' main metadata buffer
   byte padname[11]  ' filename buffer

'
'   Variables used when mounting to describe the FAT layout of the card
'   (moved to the end of the file in the Spin version).
'
'
'   Variables controlling the caching.
'
'
'  Buffering:  two sector buffers.  These two buffers must be longword
'  aligned!  To ensure this, make sure they are the first byte variables
'  defined in this object.
'
   'byte buf[SECTORSIZE] ' main data buffer
   long buf           ' 512-byte buffer which is mapped 'over' the ASM/DAT portion of this driver
   long buf2          
   long control       ' pointer to 5 longs: 2 for SD driver, 3 for Winbond
   long params

   ' From Winbond:
   long maxSize                       ' Size of Flash Memory


PUB initSDCard                       '' Initialize SD card access
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := 0
  long[control][0] := ioSpiInit << 24 | spiDO << 18 | spiClk << 12 | spiDI << 6 | spiCnt
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB stopSDCard                                         '' Stop SD card access
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := 0
  long[control][0] := ioSpiStop << 24
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB readSDCard(addr,buffer,count)                      '' Read block(s) from SD card to RAM
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := (count << 16) | (buffer & $FFFF)
  long[control][0] := (ioSpiRead << 24) | (addr & $FFFFFF)
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB writeSDCard(addr,buffer,count)                     '' Write block(s) to SD card from RAM
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := (count << 16) | (buffer & $FFFF)
  long[control][0] := (ioSpiWrite << 24) | (addr & $FFFFFF)
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB start(ctrlBlk)                                 '' Start the I2C I/O driver (standalone)
  buf := RV_getcog_overlap ' set buf as to 'overlap' HUB memory since 'sdspi' driver runs 100% in COG ram
  buf2 := buf+512

  control := ctrlBlk                                   ''  using address of 2 longs for control
  long[control][0] := 0                                ' Initialize the control block
  long[control][1] := 0

  params := @long[control][2]

'  Clr = Pin mask for select counter clear line
'  Inc = Pin mask for select counter count line
'  DIO = Pin mask for DIO or DI line (output from Propeller)
'  DO  = Pin mask for DO line (input to Propeller)
'  Clk = Pin mask for clock line
'  Returns the number of free flash blocks or 0 if none
'  Sets maxSize to the flash size in bytes or 0 if unknown JEDEC type
   ClrMask  := |< spiClr
   IncMask  := |< spiCnt
   DIOMask  := |< spiDI                  ' Initialize the I/O masks
   DOMask   := |< spiDO
   ClkMask  := |< spiClk
   outaMask := ClrMask
   diraMask := ClrMask | IncMask | DIOMask | ClkMask
   long[params][0] := long[params][1] := 0         ' Clear the parameters
   long[params][2] := invDev
  
  cognew(@i2cEntryPoint,control)            ' Start a new cog with the I2C driver
   
   sendRecv(SRAM0,WrStat|%0100_0001,0,0) ' Sequential mode, no HOLD mode
   sendRecv(SRAM1,WrStat|%0100_0001,0,0)
   maxSize := 0                        ' Default is unknown size flash
   case sendRecv(FLASH,JEDEC,3,0)      ' Get JEDEC code for the device
      JEDEC1M: maxSize := 1 * |<20     '   Atmel AT26DF081  - 1MB Flash
      JEDEC2M: maxSize := 2 * |<20     ' Winbond W25X16/16A - 2MB Flash
      JEDEC4M: maxSize := 4 * |<20     ' Winbond W25X32     - 4MB Flash
      JEDEC8M: maxSize := 8 * |<20     ' Winbond W25X64     - 8MB Flash
   
PUB RV_getcog_overlap
  return @i2cEntryPoint

PUB RV_getcog_end_of_ram
  return (end_of_cog_ram-1)<<2

PUB flashSize                          ' Return the size of the memory in bytes
   return maxSize

PUB flash_chip_erase_Start             ' RV: this command takes a couple of seconds
   sendRecv(FLASH,WrtEna,0,0)          ' Enable writes
   sendRecv(FLASH,ChipEra,0,0)     ' RV: perform full chip erase

PUB flash_chip_erase_waitDone  
   repeat until sendRecv(FLASH,Status,1,0) & Busy == 0 ' Wait until done
   
   if maxSize <> 0
      sendRecv(FLASH,WrtEna,0,0)       ' Enable writes and make sure SPRL zero
      sendRecv(FLASH,WrStat|%0000_0000,-1,0)

PRI sendRecv(d, sD, rCt, rA) | i, m    ' Send and possibly receive
'  d   = Device to be used: 1 - SRAM0, 2 - SRAM1, 3 - Flash Memory
'                           4 - ADC
'  sD  = Value to be sent MSB first, right justified in parameter
'  rCt = Number of bytes to be transferred (>0 received, <0 sent)
'  rA  = 0 or address of buffer area.  If 0, @RESULT is used
   repeat while byte[params][8] <> invDev ' Wait until done
   long[params][0] := sD                ' Possible literal value
   word[params][2] := rCt               ' Byte count
   if rA
      word[params][3] := rA
   else
      word[params][3] := @result        ' Use RESULT if rA is zero
   byte[params][8] := d                 ' Set device to use (starts)
   repeat while byte[params][8] <> invDev ' Wait until done

{
PUB reamSRAMFast0(addr, data, count)        ' Read bytes from SRAM
   repeat while byte[params][8] <> invDev ' Wait until asm fetch cog is ready
   long[params][0] := RdSRAM|addr            ' Possible literal value
   word[params][2] := count               ' Byte count
   word[params][3] := data
   byte[params][8] := SRAM0                 ' Set device to use (starts)
   ' No wait at the end: it means 'background streaming', the data is NOT available instantly!
   repeat while byte[params][8] <> invDev ' Wait until asm fetch cog is ready
}

PUB readSRAM(addr, data, count)        ' Read bytes from SRAM
'  addr  = SRAM memory starting address for reading data
'  data  = Hub starting address for data
'  count = Number of bytes to be read from SRAM memory
    sendRecv((addr>>15)+1,RdSRAM|(addr&$7FFF),count,data)

PUB readSRAM2d(addr, data, stride, skip, direction, count)        ' Read bytes from SRAM, storing data 'in 2 dimension' using a stride/skip
'  addr  = SRAM memory starting address for reading data
'  data  = Hub starting address for data
'  stride = Number of bytes stored consecutively (i.e.: 'width' of block in memory) 
'  skip  = Number of bytes in the buffer to skip before continuing read (i.e.: it corresponds to the 'pitch' - 'width' of the memory array)
'  direction = 0 for normal forward copy,  1 for reverse (horizontal flip) copy
'  count = Total number of bytes to be read from SRAM memory

'  note: data is always read LINEARLY from SRAM, but stored 2-dimensionaly in Hub RAM 
'  When using reverse copy, the copy will start at 'data + stride - 1' (in order to copy starting at end of block)
'  And 'skip' will internally be multiplied by 2 
    byte[params][9] := stride
    byte[params][10] := skip
    byte[params][11] := direction 
    sendRecv((addr>>15)+1,RdSRAM|(addr&$7FFF),count,data)
    ' Note: bytes[params][8..11] will all get cleared to 0 after SPI operation has completed

PUB readFLASH2d(addr, data, stride, skip, direction, count)        ' Read bytes from FLASH, storing data 'in 2 dimension' using a stride/skip
    ' same as 'readSRAM2d' above but from Flash storage
    byte[params][9] := stride
    byte[params][10] := skip
    byte[params][11] := direction 
    sendRecv(FLASH,RdData|addr,count,data)
    ' Note: bytes[params][8..11] will all get cleared to 0 after SPI operation has completed

PUB writeSRAM(addr, data, count)       ' Write bytes to SRAM
'  addr  = SRAM memory starting address for writing data
'  data  = Hub starting address for data or data itself
'  count = Number of bytes to be written to SRAM memory
   sendRecv((addr>>15)+1,WrSRAM|(addr&$7FFF),-count,data)

PUB readData(addr, data, count)        ' Read bytes from flash memory
'  addr  = Flash memory starting address for reading data
'  data  = Hub starting address for data
'  count = Number of bytes to be read from flash memory
    sendRecv(FLASH,RdData|addr,count,data)

PUB writeData(addr, data, count) | n ' Write bytes to flash memory
'  addr  = Flash memory starting address for writing data
'  data  = Hub starting address for data or data itself
'  count = Number of bytes to be written to flash memory
'  This routine handles the process of writing to flash memory
'  in pages of 256 bytes or less (aligned to 256 page boundaries)
   repeat while count > 0              ' Handle end of page, full pages,
      n := count <# (256 - (addr&$FF)) '  and last partial page
      sendRecv(FLASH,WrtEna,0,0)       ' Enable writes
      sendRecv(FLASH,WrData|addr,-n,data)
      repeat until sendRecv(FLASH,Status,1,0) & Busy == 0 ' Wait until done
      addr += n
      data += n                           ' Advance pointers
      count -= n

pri writeblock2(n, b)
'
'   On metadata writes, if we are updating the FAT region, also update
'   the second FAT region.
'
   writeSDCard(n, b, SECTORSIZE)
   if (n => fat1)
     if (n < fat1 + sectorsperfat)
       writeSDCard(n+sectorsperfat, b, SECTORSIZE)

pri flushifdirty
'
'   If the metadata block is dirty, write it out.
'
   if (dirty)
      writeblock2(lastread, buf2)
      dirty := 0

pri readblockc(n)
'
'   Read a block into the metadata buffer, if that block is not already
'   there.
'
   if (n <> lastread)
      flushifdirty
      readSDCard(n, buf2, SECTORSIZE)
      lastread := n

pri brword(b)
'
'   Read a byte-reversed word from a (possibly odd) address.
'
   return (byte[b]) + ((byte[b][1]) << 8)

pri brlong(b)
'
'   Read a byte-reversed long from a (possibly odd) address.
'
   return brword(b) + (brword(b+2) << 16)

pri brclust(b)
'
'   Read a cluster entry.
'
  if (filesystem == 1)
    return brword(b)
  else
    return brlong(b)

pri brwword(w, v)
'
'   Write a byte-reversed word to a (possibly odd) address, and
'   mark the metadata buffer as dirty.
'
   byte[w++] := v
   byte[w] := v >> 8
   dirty := 1

pri brwlong(w, v)
'
'   Write a byte-reversed long to a (possibly odd) address, and
'   mark the metadata buffer as dirty.
'
   brwword(w, v)
   brwword(w+2, v >> 16)

pri brwclust(w, v)
'
'   Write a cluster entry.
   if (filesystem == 1)
     brwword(w, v)
   else
     brwlong(w, v)
'
'   This may do more complicated stuff later.
'

pri getfstype : r
   if (brlong(buf+$36) == constant("F" + ("A" << 8) + ("T" << 16) + ("1" << 24)) and byte[buf][$3a]=="6")
     return 1
   if (brlong(buf+$52) == constant("F" + ("A" << 8) + ("T" << 16) + ("3" << 24)) and byte[buf][$56]=="2")
     return 2
   ' return r (default return)

pub mount : r | startBlk, sectorspercluster, reserved, rootentries, sectors
{{
'   Mount a volume.  The address passed in is passed along to the block
'   layer; see the currently used block layer for documentation.  If the
'   volume mounts, a 0 is returned, else abort is called.
}}
   if (pdate == 0)
      pdate := constant(((2009-1980) << 25) + (1 << 21) + (27 << 16) + (7 << 11))
   pclose
   stopSDCard
   if initSDCard
      abort(-19)
   lastread := -1
   dirty := 0
   readSDCard(0, buf, SECTORSIZE)
   if (getfstype > 0)
     startBlk := 0
   else
     startBlk := brlong(buf+$1c6)
     readSDCard(startBlk, buf, SECTORSIZE)
   filesystem := getfstype
   if (filesystem == 0)
      abort(-20) ' not a fat16 or fat32 volume
   if (brword(buf+$0b) <> SECTORSIZE)
      abort(-21) ' bad bytes per sector
   sectorspercluster := byte[buf][$0d]
   if (sectorspercluster & (sectorspercluster - 1))
      abort(-22) ' bad sectors per cluster
   clustershift := 0
   repeat while (sectorspercluster > 1)
      clustershift++
      sectorspercluster >>= 1
   sectorspercluster := 1 << clustershift
   clustersize := SECTORSIZE << clustershift
   reserved := brword(buf+$0e)
   if (byte[buf][$10] <> 2)
      abort(-23) ' not two FATs
   sectors := brword(buf+$13)
   if (sectors == 0)
     sectors := brlong(buf+$20)
   fat1 := startBlk + reserved
   if (filesystem == 2)
      rootentries := 16 << clustershift
      sectorsperfat := brlong(buf+$24)
      dataregion := (fat1 + 2 * sectorsperfat) - 2 * sectorspercluster
      rootdir := (dataregion + (brword(buf+$2c) << clustershift)) << SECTORSHIFT
      rootdirend := rootdir + (rootentries << DIRSHIFT)
      endofchain := $ffffff0
   else
      rootentries := brword(buf+$11)
      sectorsperfat := brword(buf+$16)
      rootdir := (fat1 + 2 * sectorsperfat) << SECTORSHIFT
      rootdirend := rootdir + (rootentries << DIRSHIFT)
      dataregion := 1 + ((rootdirend - 1) >> SECTORSHIFT) - 2 * sectorspercluster
      endofchain := $fff0
   if (brword(buf+$1fe) <> $aa55)
     abort(-24) ' bad FAT signature
   totclusters := ((sectors - dataregion + startBlk) >> clustershift)
   ' return r (default return)

pri readbytec(byteloc)
'
'   Read a byte address from the disk through the metadata buffer and
'   return a pointer to that location.
'
   readblockc(byteloc >> SECTORSHIFT)
   return buf2 + (byteloc & constant(SECTORSIZE - 1))

pri readfat(clust)
'
'   Read a fat location and return a pointer to the location of that
'   entry.
'
   fatptr := (fat1 << SECTORSHIFT) + (clust << filesystem)
   return readbytec(fatptr)

pri followchain : r
'
'   Follow the fat chain and update the writelink.
'
   r := brclust(readfat(fclust))
   writelink := fatptr
   ' return r (default return)

pri nextcluster : r
'
'   Read the next cluster and return it.  Set up writelink to
'   point to the cluster we just read, for later updating.  If the
'   cluster number is bad, return a negative number.
'
   r := followchain
   if (r < 2 or r => totclusters)
      abort(-9) ' bad cluster value
   ' return r (default return)

pri freeclusters(clust) | bp
'
'   Free an entire cluster chain.  Used by remove and by overwrite.
'   Assumes the pointer has already been cleared/set to end of chain.
'
   repeat while (clust < endofchain)
      if (clust < 2)
         abort(-26) ' bad cluster number")
      bp := readfat(clust)
      clust := brclust(bp)
      brwclust(bp, 0)
   flushifdirty

pri datablock
'
'   Calculate the block address of the current data location.
'
   return (fclust << clustershift) + dataregion + ((floc >> SECTORSHIFT) & ((1 << clustershift) - 1))


pri pflushbuf(rcnt, metadata) : r | cluststart, newcluster, count, i
'
'   Flush the current buffer, if we are open for write.  This may
'   allocate a new cluster if needed.  If metadata is true, the
'   metadata is written through to disk including any FAT cluster
'   allocations and also the file size in the directory entry.
'
   if (direntry == 0)
      abort(-27) ' not open for writing
   if (rcnt > 0) ' must *not* allocate cluster if flushing an empty buffer
      if (frem < SECTORSIZE)
         ' find a new clustercould be anywhere!  If possible, stay on the
         ' same page used for the last cluster.
         newcluster := -1
         cluststart := fclust & (!((SECTORSIZE >> filesystem) - 1))
         count := 2
         repeat
            readfat(cluststart)
            repeat i from 0 to SECTORSIZE - 1<<filesystem step 1<<filesystem
              if (byte[buf2][i] == 0)
                if (brclust(buf2+i) == 0)
                   newcluster := cluststart + (i >> filesystem)
                   if (newcluster => totclusters)
                      newcluster := -1
                   quit
            if (newcluster > 1)
               brwclust(buf2+i, endofchain+$f)
               if (writelink == 0)
                  brwword(readbytec(direntry)+$1a, newcluster)
                  writelink := (direntry&(SECTORSIZE-filesystem))
                  brwlong(buf2+writelink+$1c, floc+bufat)
                  if (filesystem == 2)
                     brwword(buf2+writelink+$14, newcluster>>16)
               else
                  brwclust(readbytec(writelink), newcluster)
               writelink := fatptr + i
               fclust := newcluster
               frem := clustersize
               quit
            else
               cluststart += (SECTORSIZE >> filesystem)
               if (cluststart => totclusters)
                  cluststart := 0
                  count--
                  if (rcnt < 0)
                     rcnt := -5 ' No space left on device
                     quit
      if (frem => SECTORSIZE)
         writeSDCard(datablock, buf, SECTORSIZE)
         if (rcnt == SECTORSIZE) ' full buffer, clear it
            floc += rcnt
            frem -= rcnt
            bufat := 0
            bufend := rcnt
   if (rcnt < 0 or metadata) ' update metadata even if error
      readblockc(direntry >> SECTORSHIFT) ' flushes unwritten FAT too
      brwlong(buf2+(direntry & (SECTORSIZE-filesystem))+$1c, floc+bufat)
      flushifdirty
   if (rcnt < 0)
      abort(rcnt)
   return rcnt

pub pflush
{{
'   Call flush with the current data buffer location, and the flush
'   metadata flag set.
}}
   return pflushbuf(bufat, 1)

pri pfillbuf : r
'
'   Get some data into an empty buffer.  If no more data is available,
'   return -1.  Otherwise return the number of bytes read into the
'   buffer.
'
   if (floc => filesize)
      return -1
   if (frem == 0)
      fclust := nextcluster
      frem := (clustersize) <# (filesize - floc)
   readSDCard(datablock, buf, SECTORSIZE)
   r := SECTORSIZE
   if (floc + r => filesize)
      r := filesize - floc
   floc += r
   frem -= r
   bufat := 0
   bufend := r
   ' return r (default return)

pub pclose : r
{{
'   Flush and close the currently open file if any.  Also reset the
'   pointers to valid values.  If there is no error, 0 will be returned.
}}
   if (direntry)
      r := pflush
   bufat := 0
   bufend := 0
   filesize := 0
   floc := 0
   frem := 0
   writelink := 0
   direntry := 0
   fclust := 0
   firstcluster := 0
   ' return r (default return)

pub setdate(year, month, day, hour, minute, second)
{{
'   Set the current date and time, as a long, in the format
'   required by FAT16.  Various limits are not checked.
}}
   pdate := ((year-1980) << 25) + (month << 21) + (day << 16)
   pdate += (hour << 11) + (minute << 5) + (second >> 1)

pub popen(s, mode) : r | i, sentinel, dirptr, freeentry
{{
'   Close any currently open file, and open a new one with the given
'   file name and mode.  Mode can be "r" "w" or "d" (delete).
'   If the file is opened successfully, 0 will be returned.  If the
'   file did not exist, and the mode was not "w", -1 will be
'   returned.  Otherwise abort will be called with a negative error
'   code.
}}
   pclose
   i := 0
   repeat while (i<8 and byte[s] and byte[s] <> ".")
      padname[i++] := byte[s++]
   repeat while (i<8)
      padname[i++] := " "
   repeat while (byte[s] and byte[s] <> ".")
      s++
   if (byte[s] == ".")
      s++
   repeat while (i<11 and byte[s])
      padname[i++] := byte[s++]
   repeat while (i < 11)
      padname[i++] := " "
   sentinel := 0
   freeentry := 0
   repeat dirptr from rootdir to rootdirend - DIRSIZE step DIRSIZE
      s := readbytec(dirptr)
      if (freeentry == 0 and (byte[s] == 0 or byte[s] == $e5))
         freeentry := dirptr
      if (byte[s] == 0)
         sentinel := dirptr
         quit
      repeat i from 0 to 10
         if (padname[i] <> byte[s][i])
            quit
      if (i == 11 and 0 == (byte[s][$0b] & $18)) ' this always returns
         fclust := brword(s+$1a)
         if (filesystem == 2)
            fclust += brword(s+$14) << 16
         firstcluster := fclust
         filesize := brlong(s+$1c)
         if (mode == "r")
            frem := (clustersize) <# (filesize)
            return 0
         if (byte[s][11] & $d9)
            abort(-6) ' no permission to write
         if (mode == "d")
            brwword(s, $e5)
            if (fclust)
               freeclusters(fclust)
            flushifdirty
            return 0
         if (mode == "w")
            brwword(s+$1a, 0)
            brwword(s+$14, 0)
            brwlong(s+$1c, 0)
            writelink := 0
            direntry := dirptr
            if (fclust)
               freeclusters(fclust)
            bufend := SECTORSIZE
            fclust := 0
            filesize := 0
            frem := 0
            return 0
         abort(-3) ' bad argument
   if (mode <> "w")
      return -1 ' not found
   direntry := freeentry
   if (direntry == 0)
      abort(-2) ' no empty directory entry
   ' write (or new append): create valid directory entry
   s := readbytec(direntry)
   bytefill(s, 0, DIRSIZE)
   bytemove(s, @padname, 11)
   brwword(s+$1a, 0)
   brwword(s+$14, 0)
   i := pdate
   brwlong(s+$e, i) ' write create time and date
   brwlong(s+$16, i) ' write last modified date and time
   if (direntry == sentinel and direntry + DIRSIZE < rootdirend)
      brwword(readbytec(direntry+DIRSIZE), 0)
   flushifdirty
   writelink := 0
   fclust := 0
   bufend := SECTORSIZE
   ' return r (default return)

pub pread(ubuf, count) : r | t
{{
'   Read count bytes into the buffer ubuf.  Returns the number of bytes
'   successfully read, or a negative number if there is an error.
'   The buffer may be as large as you want.
}}
   repeat while (count > 0)
      if (bufat => bufend)
         t := pfillbuf
         if (t =< 0)
            if (r > 0)
' parens below prevent this from being optimized out
              return (r)
            return t
      t := (bufend - bufat) <# (count)
      if ((t | (ubuf) | bufat) & 3)
         bytemove(ubuf, buf+bufat, t)
      else
         longmove(ubuf, buf+bufat, t>>2)
      bufat += t
      r += t
      ubuf += t
      count -= t
   ' return r (default return)

pub pwrite(ubuf, count) : r | t
{{
'   Write count bytes from the buffer ubuf.  Returns the number of bytes
'   successfully written, or a negative number if there is an error.
'   The buffer may be as large as you want.
}}
   repeat while (count > 0)
      if (bufat => bufend)
         pflushbuf(bufat, 0)
      t := (bufend - bufat) <# (count)
      if ((t | (ubuf) | bufat) & 3)
         bytemove(buf+bufat, ubuf, t)
      else
         longmove(buf+bufat, ubuf, t>>2)
      r += t
      bufat += t
      ubuf += t
      count -= t
   ' return r (default return)
{{
'   Write a null-terminated string to the file.
}}

pub seek(pos) | delta
   if (direntry or pos < 0 or pos > filesize)
      return -1

   ' rv fix tentative:
  fclust := firstcluster
  frem := (clustersize) <# (filesize)
  floc := 0
  bufat := 0
  bufend := 0
  delta := 0
      
   repeat while (pos => delta + clustersize)
      fclust := nextcluster
      delta += clustersize
      
   floc := delta
   frem := (clustersize) <# (filesize - floc)

   if (bufend == 0 or pos < floc - bufend or pos => floc - bufend + SECTORSIZE)
      ' must change buffer
      delta := floc + frem
      floc := pos & - SECTORSIZE
      frem := delta - floc
      pfillbuf
   bufat := pos & (SECTORSIZE - 1)
   return 0

{
pub get_filesize
   return filesize

pub tell
   return floc + bufat - bufend

pub getclustersize
  return clustersize

pub getclustercount
  return totclusters
}

'' System Version 2.003.rr001
'' loaderInit Changes:
''
''   2006/09/24 - Corrected action table for i2cRead0Cmd
''   2006/10/02 - Improved assembly comments.  Changed i2cReset, i2cStop
''   2006/11/02 - Changed data setup time.  Changed i2cReset timing
''   2006/11/03 - Changed read/write method speed to 100KHz
''   2006/11/04 - Added checkPresence, writeWait, and computeTimes methods
''   2006/11/06 - Limited boot loading to 32K - 16 (OS uses last 16 bytes)
''                Added ioVerifyCmd and verifyEEPROM method
''   2006/11/09 - Modified boot and verify to use the minimum of the actual
''                length of the program stored in vbase ($0008) or the
''                specified byte count.
''                Note also that these routines don't know about option bits.
''   2006/11/10 - Now control block address is passed to start routine
''   2007/01/09 - Added SPI routines for SD card FAT access
''   2007/01/13 - Corrected checksum when ioBoot or ioVerify
''   2007/01/13 - Stores stack marker & clears VAR area on ioBootCmd
''   2007/02/06 - Stores stack marker & clears VAR area on ioSpiBoot
''   2007/02/13 - Changed the way verify mode was done, combined code
''   2007/02/20 - Added ioStopLdr option to stop the loader's cog
''   2007/02/22 - Corrected bootSDCard.  Needs start and initSDCard calls first.
''   2009/04/02 - Cluso99: Fix to force the SD card to tristate DO in spiDoStop
''   2010/10/27 - Remi Veilleux: Code cleanup: delete every unneeded item and tries to use less memory


'' This portion is normally run only once during initialization and the driver remains
'' resident in a cog.  These routines can be used completely independently of the rest
'' of the Propeller OS.  The start routine here expects the address of a 2 long area
'' to be used for passing information to the I/O routines in the COG.  This area should
'' be located in an area of memory not expected to be overlaid by data or a program that
'' might be loaded since the COG routines will be accessing this information after an
'' operation has completed.

'' This object provides an I2C EEPROM read/write routine that can handle both 100KHz and
'' 400KHz bus speeds and EEPROM page sizes of 64, 128, or 256 bytes (or no paging/no delay
'' as with Ramtron serial RAM).  The SPIN interpreter can be started after reading, either
'' in the same COG used by these routines or in a free COG.  The control information is
'' passed in a 2 long parameter block whose address is passed to the COG when it is started.
'' The parameter block is updated when the operation is completed.  Note that these are shown
'' here as they appear in a long value rather than the order of the bytes in memory.

'' -------------------------------------------------------------------
'' |   cmd/status   |          I/O pin / device / address            |
'' -------------------------------------------------------------------
'' |           byte count           |          HUB address           |
'' -------------------------------------------------------------------

'' The EEPROM address is in the same format used by other routines with the I/O pin pair
'' in bits 21..19, the device address in bits 18..16, and the 64K address in bits 15..0.
'' Note that the I/O pin pair is the number of the SCL pin divided by 2.  The SDA pin is
'' always the next higher numbered pin.  The command code is in the low order bits of the
'' high order byte of the first long (see ioCmdMask).  This is always non-zero to indicate
'' that a command is to be performed by the COG routines.  When the command is finished,
'' this is set to zero.  The errorFlag bit is set to one if a NAK was read after a write
'' transfer.  This is the only error reported by these routines.  A read operation and
'' zero-length writes do involve several write transfers for addressing, but the data
'' read transfer has no error checking.  When the command is completed, the device address,
'' byte count, and HUB address are all updated to their values at that time.  For the
'' verify operation (ioVerifyCmd), an error is reported if the checksum is not zero and
'' the HUB address field is not incremented.  It may be used for some other checksum
'' reporting in the future.

'' The pins used for the boot EEPROM I2C bus (at least on Parallax's Demo Board) do not
'' have a pullup on SCL.  This requires that SCL be driven both high and low.  If the bus
'' used is on pins 28 and 29, SCL is actively driven at all times.

'' These EEPROM read/write routines do not provide for waiting for the write to complete
'' nor do they check for paged writes.  All bytes in a multi-byte write must lie within
'' a single EEPROM page since the EEPROM write address counter wraps around at a page
'' boundary.  Similarly, for multi-byte reads, all requested bytes must lie within the
'' same device since the sequential read counter wraps around at the device boundary.

'' Command codes are provided for devices with zero, one, or two address bytes following
'' the device selection byte.  As for all I2C devices, addressing is done using write
'' mode and the device is reselected in read mode after the last address byte.  In the
'' case of ioRead0Cmd, the device is initially selected in read mode.  For 8-bit addresses,
'' the device select code is taken from bits 15-8 of the address value.  For the case
'' without address bytes, the device select code is taken from bits 7-0 of the address value.
'' These device select codes must have their least significant bit set to zero (for write
'' mode) except in the case of ioRead0Cmd where it must be set to one for proper operation.

'' SPI data is handled a little differently.  For ioSpiInit, the 6 bit pin numbers for DO,
'' Clk, DI, and CS are given from MSB to LSB of the 24 bit address field of the command and
'' are used for all further I/O operations (until an ioSpiStop is done).

DAT
                        org     0
i2cEntryPoint           ' Entrypoint / loop point for merged Winbond + SD driver

                        ' First part: Winbond
                        ' Check if a command has been requested for SRAM/FLASH portion
entryPoint              mov     dira, #0        ' RV: clear all pins back to 'input' to prevent conflict with other cogs
                        mov     outa, #0        ' Clear outa 
                        mov     flashTemp,PAR
                        add     flashTemp,#8   ' Offset to start of Winbond params (PAR + 2 longs) 
                        rdlong  flashpreamble+0,flashTemp    ' Get parameters
                        add     flashTemp,#4
                        rdword  flashpreamble+1,flashTemp    ' Byte count
                        add     flashTemp,#2
                        test    flashpreamble+1,startROM  wc ' Extend word sign
                        rdword  flashpreamble+2,flashTemp    ' Starting address
                 if_c   or      flashpreamble+1,wordSign
                        add     flashTemp,#2
                        rdlong  flashpreamble+3,flashTemp  ' Device # and additional params
                        mov     flashTemp,flashpreamble+3
                        and     flashTemp, #255         wz  ' extract device # byte
                 if_z   jmp     #sddriver_start             ' If null device, go back and wait
                        or      outa,outaMask
                        or      dira,diraMask           ' DOpin is only input
                        andn    outa,ClrMask
                        nop                             ' Allow extra time
                        or      outa,ClrMask            ' Clear the select count
                        nop                             ' Allow extra time
:selectDev              or      outa,IncMask
                        nop                             ' Allow extra time
                        andn    outa,IncMask            ' Increment select count
                        djnz    flashTemp,#:selectDev
                        mov     flashMask,#$FF          ' Set up sending mask
                        shl     flashMask,#24           ' m := $FF000000
:scanMask               test    flashpreamble+0,flashMask wz ' sD & m == 0?
                 if_z   shr     flashMask,#8            '  m >>= 8, m == 0?
                 if_z   tjnz    flashMask,#:scanMask
                        and     flashMask,msbMask       ' m &= $80808080
                        tjz     flashMask,#:skipImm     ' Skip if nothing
:sendImm                test    flashpreamble+0,flashMask wz ' Output bit to be sent
                        muxnz   outa,DIOMask
                        or      outa,ClkMask            ' Toggle clock
                        shr     flashMask,#1            ' Advance bit mask here
                        andn    outa,ClkMask            '  to allow extra time
                        tjnz    flashMask,#:sendImm     ' Continue to lsb
:skipImm                test    flashpreamble+1,cmdMask   wc ' If byte count < 0
                if_nc   jmp     #skipSend               '  then transmit data
                        neg     flashpreamble+1,flashpreamble+1   ' Make transmit count
sendData                mov     flashMask,#$80          '  positive
                        rdbyte  flashTemp,flashpreamble+2    ' Get the data byte and
                        add     flashpreamble+2,#1           '  increment address
:bitLoop                test    flashTemp,flashMask  wz ' Output bit to be sent
                        muxnz   outa,DIOMask
                        or      outa,ClkMask            ' Toggle clock
                        shr     flashMask,#1            ' Advance bit mask here
                        andn    outa,ClkMask            '  to allow extra time
                        tjnz    flashMask,#:bitLoop     ' Continue to lsb
                        djnz    flashpreamble+1,#sendData    ' Continue to next byte
skipSend                andn    dira,DIOMask            ' Don't need DIO now
                        tjz     flashpreamble+1,#skipRecv    ' Skip if count == 0

                        ' RV: test unrolled fast read
                        ' Maximum throughput is: 4 inst per bit = 4*8 = 32 inst per byte + approx 4
                        ' Total: approx 36 inst = 144 clocks per byte
                        ' at 80mhz = 80000000 / 144 =  555555 bytes per second
                        ' Which means 9259 bytes per frame, which equals at about 243 raster per screen = 38 bytes per raster line
                        ' In my test: 768 bytes in about 26 raster line = 29.5 bytes per raster = 7177 bytes per frame = 430670 bytes per second
                        ' We are running at about 77% of the potential max speed

                        ' RV: add support for '2D' memory read using a stride and skip byte value
                        ' TODO: add 'horizontal flip' copy
                        mov     readStride,flashpreamble+3 ' fetch params long
                        shr     readStride, #8
                        and     readStride, #255         ' Extract stride
                        mov     readStrideCount, readStride
                        mov     readSkip,flashpreamble+3 ' fetch params long
                        shr     readSkip, #16
                        and     readSkip, #255         ' Extract skip
                        mov     readDirection,flashpreamble+3 ' fetch params long
                        shr     readDirection, #24
                        and     readDirection, #255 wz   ' Extract direction
              if_z      mov    self_mod_add, self_mod_direction_add  
              if_z      jmp     #recvByte               ' if forward copy, nothing special to do: start copying
               
                        ' Reverse copy: perform various adjustment:
                        ' Start of reverse copy is at the end of the stride - 1
                        add     flashpreamble+2,readStride
                        sub     flashpreamble+2,#1
                        ' Adjust skip = skip + stride*2 to give a correct skip to next line
                        mov     flashTemp, readStride
                        shl     flashTemp, #1
                        add     readSkip, flashTemp
                        ' Set sub instead of add:
                        mov    self_mod_add, self_mod_direction_sub
recvByte                
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<7     ' Copy to result byte
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<6     ' Copy to result byte
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<5     ' Copy to result byte
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<4     ' Copy to result byte
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<3     ' Copy to result byte
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<2     ' Copy to result byte
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<1     ' Copy to result byte
                        or      outa,ClkMask            ' Toggle clock
                        test    DOMask,ina           wz ' Test input bit
                        andn    outa,ClkMask
                        muxnz   flashTemp,#1<<0     ' Copy to result byte

                        wrbyte  flashTemp,flashpreamble+2    ' Store the data byte
self_mod_add            add     flashpreamble+2,#1           ' Increment or Decrement address

                        ' RV 2D memory support: decrement readStrideCount: once reach zero, advance memory address then resume
                        ' Note: if 'readSkip' is 0, then these 3 lines have absolutely no effect, whatever 'readStrideCount' is.
                        sub     readStrideCount, #1 wz         ' Decrement stride count
              if_z      add     flashpreamble+2,readSkip       ' Skip memory address according to skip value
              if_z      mov     readStrideCount, readStride    ' Reset stridecount for next row
              
                        djnz    flashpreamble+1,#recvByte    ' Continue to next byte

skipRecv                andn    outa,ClrMask            ' Done. Deselect SD card
                        andn    dira,ClkMask            ' Don't need Clk now
                        mov     flashTemp,PAR           ' Force null device
                        add     flashTemp,#8+8         '  to notify sendRecv
                        wrlong  zeroVal,flashTemp       '   that we're done
                        jmp     #entryPoint

sddriver_start          ' Start point for SD driver
i2cNewOpFetch           rdlong  i2cAddr,PAR             ' Fetch control information
                        mov     i2cCmd,i2cAddr
                        shr     i2cCmd,#24              ' Isolate command code
                        mov     Options,i2cCmd
                        and     i2cAddr,i2cAddrMask     ' Only need address at this point
                        and     i2cCmd,#ioCmdMask  wz
                if_z    jmp     #i2cEntryPoint          ' Wait for a new operation
                        mov     i2cTemp,PAR
                        add     i2cTemp,#4              ' Now get 2nd long of packet
                        rdlong  i2cCount,i2cTemp
                        mov     i2cBufAdr,i2cCount      ' Byte count
                        rdlong  SaveClkFreq,#clkfreqVal ' Save clock frequency and mode
                        shr     i2cCount,#16
                        and     i2cBufAdr,i2cWordMask   ' HUB RAM address of buffer
                        movs    ShiftData,#0            ' Initialize for saving Preamble
                        mov     StoreLocal,initStore    '  on I2C and SPI reads
                        mov     Preamble+0,#0
                        mov     Preamble+1,#0
                        mov     Preamble+2,#0
                        mov     Preamble+3,#0
                        test    Options,#ioNoStore wc
                        test    i2cCmd,#ioSpiMask  wz   ' Check for SPI commands
                if_nz   jmp     #spiEntryPoint

' betz -- get rid of the i2c stuff to save space

                        jmp     #i2cEntryPoint
                        
'' SPI routines for Rokicki's SD card FAT file system driver

spiEntryPoint
                        mov     dira, dira_backup       ' RV: restore pins to prevent conflict with other cogs
                        mov     outa, outa_backup
                        test    i2cCmd,#ioBootCmd  wc   ' Check for boot
              if_c      jmp     #spiDoRead              ' (Treat like read)
                        cmp     i2cCmd,#ioSpiStop wc,wz
              if_c      jmp     #spiDoInit              ' Decode operation
              if_z      jmp     #spiDoStop
                        cmp     i2cCmd,#ioSpiWrite wc
              if_c      jmp     #spiDoRead
                        jmp     #spiDoWrite

'' Initialize SPI communications.  The pin numbers of the 4 I/O pins are
'' provided in the 24 bit address field of the control packet.  From MSB to
'' LSB, these are DO - Data Out, Clk - Clock, DI - Data In, CS - Card Select.

spiDoInit               movd    :moveIt,#spiMaskInc
                        mov     spiBlkCnt,#4
:makeMask               mov     i2cMask,#1
                        mov     i2cTemp,i2cAddr         ' Only use lower 5 bits of
                        and     i2cTemp,#%11111         '  6 bit shift count field
                        shl     i2cMask,i2cTemp
:moveIt                 mov     0-0,i2cMask             ' Store the bit mask for the pin
                        cmp     spiBlkCnt,#1       wz
              if_ne     or      outa,i2cMask            ' Make all pins high outputs
              if_ne     or      dira,i2cMask            '  except DO is an input since
              if_e      andn    dira,i2cMask            '   input/output is card relative
                        sub     :moveIt,incrDst
                        ror     i2cAddr,#6
                        djnz    spiBlkCnt,#:makeMask
                        or      outa,spiMaskClr         ' Fix this!!!!
                        or      dira,spiMaskClr         ' Fix this!!!!
                        andn    outa,spiMaskInc         ' Fix this!!!!
                        'or      outa,ledMask           ' Betz - debug
                        'or      dira,ledMask           ' Betz - debug
                        rol     i2cAddr,#24             ' Leave i2cAddr unchanged
                        mov     i2cTime,cnt             ' Set up a 1 second timeout
                        call    #deselectSD
                        mov     spiBlkCnt,spiInitCnt
:initRead               call    #spiRecvByte            ' Output a stream of 32K clocks
                        djnz    spiBlkCnt,#:initRead    '  in case SD card left in some
                        call    #selectSD
                        mov     spiOp,#0                '   undefined state
                        mov     spiParm,#0
                        call    #spiSendCmd             ' Send a reset command and deselect
:waitIdle               mov     spiOp,#55
                        call    #spiSendCmd             ' APP_CMD (Application Specific)
                        mov     spiOp,#41
                        call    #spiSendCmd             ' SEND_OP_COND (Initialization)
                        cmp     i2cData,#1         wz   ' Wait until response not In Idle
              if_e      jmp     #:waitIdle
                        call    #deselectSD
                        tjz     i2cData,#i2cGoUpdate    ' Initialization complete
                        or      i2cAddr,errorFlag
                        jmp     #i2cGoUpdate            ' Could not initialize the card

' Modification by Cluso99 to force the SD card to tristate the D0 pin (allows other chips on the bus)
spiDoStop               call    #deselectSD             ' CS=1
                        andn    dira,spiMaskDO          ' disable D0 as output (s/be input anyway)
                        call    #spiRecvByte            ' Output a stream of clocks 
                        andn    dira,spiMaskClr         ' Clr input
                        'mov     spiMaskClr,#0          ' Fix this!!! Clr is not initialized
                        andn    dira,spiMaskInc         ' Inc input
                        mov     spiMaskInc,#0
                        mov     spiMaskDO,#0            ' D0 already input (above)
                        andn    dira,spiMaskDI          ' DI input
                        mov     spiMaskDI,#0            '  
                        andn    dira,spiMaskClk         ' CLK input
                        mov     spiMaskClk,#0
                        jmp     #i2cGoUpdate


'' Read one or more 512 byte blocks and store the specified number of bytes
'' into the HUB location given.  The block number is provided in the 24 bit
'' address field and incremented after every block is read.  Partial blocks are
'' allowed and any extra bytes read are discarded.

spiDoRead               call    #selectSD               ' Read from specified block
                        mov     spiOp,#17               ' READ_SINGLE_BLOCK
:readRepeat             mov     i2cTime,cnt             ' Save start of timeout
                        mov     spiParm,i2cAddr
                        call    #spiSendCmd             ' Read from specified block
                        call    #spiResponse
                        mov     spiBlkCnt,spiBlkSize    ' Transfer a block at a time
:getRead                call    #spiRecvByte
                        tjz     i2cCount,#:skipStore    ' Check for count exhausted
                        call    #StoreData
                        sub     i2cCount,#1
:skipStore              djnz    spiBlkCnt,#:getRead     ' Are we done with the block?
                        call    #spiRecvByte
                        call    #spiRecvByte            ' Yes, finish with 16 clocks
                        add     i2cAddr,#1
                        tjnz    i2cCount,#:readRepeat   '  and check for more blocks to do
                        call    #deselectSD             ' Increment address, deselect card
checkEndIO              test    i2cCmd,#ioBootCmd  wc
                 if_nc  jmp     #i2cGoUpdate            ' If not booting, we're done

stopThisCOG             jmp     #stopThisCOG
                      
'' Write one or more 512 byte blocks with the specified number of bytes from
'' the HUB location given.  The block number is provided in the 24 bit address
'' field and incremented after every block is written.  Partial blocks are
'' allowed and are padded with zeroes.

spiDoWrite              call    #selectSD               ' Write to specified block
:writeRepeat            mov     spiOp,#24               ' WRITE_BLOCK
                        mov     i2cTime,cnt             ' Setup timeout
                        mov     spiParm,i2cAddr
                        call    #spiSendCmd             ' Write to specified block
                        mov     i2cData,#$FE            ' Ask to start data transfer
                        call    #spiSendByte
                        mov     spiBlkCnt,spiBlkSize    ' Transfer a block at a time
:putWrite               mov     i2cData,#0              '  padding with zeroes if needed
                        tjz     i2cCount,#:padWrite     ' Check for count exhausted
                        rdbyte  i2cData,i2cBufAdr       ' If not, get the next data byte
                        add     i2cBufAdr,#1
                        sub     i2cCount,#1
:padWrite               call    #spiSendByte
                        djnz    spiBlkCnt,#:putWrite    ' Are we done with the block?
                        call    #spiRecvByte
                        call    #spiRecvByte            ' Yes, finish with 16 clocks
                        call    #spiResponse
                        and     i2cData,#$1F            ' Check the response status
                        cmp     i2cData,#5         wz
              if_ne     or      i2cAddr,errorFlag       ' Must be Data Accepted
              if_ne     jmp     #i2cGoUpdate
                        movs    spiWaitData,#0          ' Wait until not busy
                        call    #spiWaitBusy
                        add     i2cAddr,#1
                        tjnz    i2cCount,#:writeRepeat  '  to next if more data remains
                        call    #deselectSD             ' Increment block address and go
                        jmp     #i2cGoUpdate

'' Mid level SPI I/O

spiSendCmd              call    #spiRecvByte            '  selecting card and clocking
                        mov     i2cData,spiOp
                        or      i2cData,#$40            ' Send command byte (1st 2 bits %01)
                        call    #spiSendByte
                        mov     i2cData,spiParm
                        shr     i2cData,#15             ' Supplied address is sector number
                        call    #spiSendByte 
                        mov     i2cData,spiParm         ' Send to SD card as byte address,
                        shr     i2cData,#7              '  in multiples of 512 bytes
                        call    #spiSendByte
                        mov     i2cData,spiParm         ' Total length of this address is
                        shl     i2cData,#1              '  four bytes
                        call    #spiSendByte
                        mov     i2cData,#0
                        call    #spiSendByte
                        mov     i2cData,#$95            ' CRC code (for 1st command only)
                        call    #spiSendByte
spiResponse             movs    spiWaitData,#$FF        ' Wait for response from card
spiWaitBusy             call    #spiRecvByte
                        mov     i2cTemp,cnt
                        sub     i2cTemp,i2cTime         ' Check for expired timeout (1 sec)
                        cmp     i2cTemp,SaveClkFreq wc
              if_nc     or      i2cAddr,errorFlag
              if_nc     jmp     #i2cGoUpdate
spiWaitData             cmp     i2cData,#0-0       wz   ' Wait for some other response
              if_e      jmp     #spiWaitBusy            '  than that specified
spiSendCmd_ret
spiResponse_ret
spiWaitBusy_ret         ret

'' Low level byte I/O

spiSendByte             mov     i2cMask,#%10000000
:sendBit                test    i2cMask,i2cData    wc
                        andn    outa,spiMaskClk         ' Send data bytes MSB first
                        muxc    outa,spiMaskDI
                        or      outa,spiMaskClk
                        shr     i2cMask,#1              ' When mask shifted out, we're done
                        tjnz    i2cMask,#:sendBit
                        or      outa,spiMaskDI          ' Leave DI in idle (high) state
spiSendByte_ret         ret

spiRecvByte             mov     i2cMask,#%10000000
:recvBit                andn    outa,spiMaskClk         ' Receive data bytes MSB first
                        or      outa,spiMaskClk         ' Copy DO to data bit
                        test    spiMaskDO,ina      wc     
                        muxc    i2cData,i2cMask
                        shr     i2cMask,#1              ' When mask shifted out, we're done
                        tjnz    i2cMask,#:recvBit
                        and     i2cData,#%11111111      ' Eight bits received
spiRecvByte_ret         ret

selectSD                call    #deselectSD
                        mov     i2cTemp,#spiSel
:loop                   or      outa,spiMaskInc
                        andn    outa,spiMaskInc
                        djnz    i2cTemp,#:loop
                        'andn    outa,ledMask
selectSD_ret            ret

deselectSD              andn    outa,spiMaskClr
                        or      outa,spiMaskClr
                        'or      outa,ledMask
deselectSD_ret          ret

'' For both I2C and SPI, store data on a read operation unless ioNoStore is set.
'' Accumulate a checksum and always save a copy of the first 16 bytes read.
'' If this is an ioBootCmd or ioSpiBoot, adjust the amount to be read based
'' on the value in the program preamble in the word at vbase ($0008).

StoreData               test    Options,#ioNoStore wc
                if_nc   wrbyte  i2cData,i2cBufAdr       ' Store data in specified location
                        add     i2cBufAdr,#1            '  and increment the address
ShiftData               shl     i2cData,#0-0
StoreLocal              or      Preamble+0,i2cData      ' Store a local copy of the program
                        add     ShiftData,#8            '  preamble for when we're reading
                        cmp     ShiftData,testIns  wz   '   in a new Spin program
                if_z    movs    ShiftData,#0            ' Pack the data into successive longs
                if_z    add     StoreLocal,incrDst
                if_z    cmp     StoreLocal,testDst wz   ' Stop after saving $0010 bytes
                if_z    mov     StoreLocal,noStore
                if_z    test    i2cCmd,#ioBootCmd  wc   ' If we're reading in a new program,
          if_c_and_z    mov     i2cCount,Preamble+2     '  change i2cCount to vbase adjusted
          if_c_and_z    and     i2cCount,i2cWordMask    '  by number of bytes loaded so far.
          if_c_and_z    sub     i2cCount,#16 - 1        ' i2cCount will be decremented again
StoreData_ret           ret

'' The operation has completed, with or without errors.  Update the control block
'' in main memory and wait for the next operation to be requested.

i2cGoUpdate             and     i2cBufAdr,i2cWordMask   ' Copy updated information
                        shl     i2cCount,#16            '  back to control packet
                        or      i2cCount,i2cBufAdr
                        mov     i2cTemp,PAR
                        add     i2cTemp,#4
                        wrlong  i2cCount,i2cTemp
                        ' RV: backup 'dira' and clear all output pins
                        mov     dira_backup, dira
                        mov     dira, #0
                        mov     outa_backup, outa
                        mov     outa, #0
                        wrlong  i2cAddr,PAR             ' Indicate operation is done
                        
                        jmp     #i2cEntryPoint          '  and go wait for a new one

self_mod_direction_add  add     flashpreamble+2,#1               
self_mod_direction_sub  sub     flashpreamble+2,#1

'' This action table contains bit sequences for controlling device addressing and read/write
'' mode selection for each of the commands possible.  From LSB to MSB, the actions are:
'' 0 - Write the EEPROM device select code for write mode and 2 address bytes
'' 1 - Write the MSB device address or (for ioRead1Cmd/ioWrite1Cmd) a device select code
'' 2 - Write the LSB device address or (for ioRead0Cmd/ioWrite0Cmd) a device select code
'' 3 - Output a Start Sequence prior to reselecting in read mode
'' 4 - Fetch a data value for writing
'' 5 - Construct an EEPROM device select code for read mode and 2 address bytes
'' 6 - Construct a read mode device select code from the MSB of the 16 bit device address
'' 7 - Write the data value or read mode device select code
'' 8 - Read a byte of data from the device and store it

'' Constants for all routines

i2cWordMask             long    $0000FFFF
i2cAddrMask             long    $00FFFFFF
errorFlag               long    $80000000               ' NAK received during write cycle
spiBlkSize                                              ' Number of bytes in an SD card block
incrDst                 long    %10_00000000            ' Used to increment destination field
testIns                 shl     i2cData,#32             ' Used to compare for end of word packing
initStore               or      Preamble+0,i2cData      ' Used to initialize packing instruction
testDst                 or      Preamble+4,i2cData      ' Used to check for end of packing buffer
noStore                 jmp     #StoreData_ret          ' Used after all data stored into Preamble
spiInitCnt              long    32768 / 8               ' Initial SPI clocks produced

'' Variables for all routines
spiMaskClr              long    1<<spiClr
'ledMask                 long    1<<15

size                    long spiMaskClr

' From Winbond:
zeroVal                 long    0                       ' Zero value
cmdMask                 long    $80000000               ' Also used for long sign
msbMask                 long    $80808080               ' MSB mask for reading
wordSign                long    $FFFF0000               ' For extending word sign
startROM                long    $8000                   ' Also used for word sign
diraMask                long    0               ' Clr, Clk, DIO, Inc
outaMask                long    0               ' /Clear select counter
ClrMask                 long    0               ' /Clear select counter
IncMask                 long    0              ' Increment counter
DOMask                  long    0               ' DO from flash or SRAM
DIOMask                 long    0               ' DIO to flash or SRAM
ClkMask                 long    0               ' Clock to flash or SRAM

padding                 long    0,0             ' Padding to allow more variables in gfx_buffer

end_of_cog_ram          long end_of_cog_ram

Preamble                res 4              ' Private copy of program preamble
i2cCmd                  res 1
i2cTemp                 res 1
i2cCount                res 1
i2cBufAdr               res 1
i2cAddr                 res 1
SaveClkFreq             res 1                       ' Initial clock frequency (clkfreqVal)
spiBlkCnt               res 1                       ' Number of SD card bytes to go in block
Options                 res 1                       ' Option bits (ioNoStore, ioLowSpeed)

'' Local variables for low level I2C routines

spiOp                   res 1                           ' Operation code for SPI command
spiParm                 res 1                           ' Parameter value for SPI command
i2cTime                 res 1                        ' Used for timekeeping
i2cData                 res 1                        ' Data to be transmitted / received
i2cMask                 res 1                          ' Bit mask for bit to be tx / rx

'' Additional local variables for SPI SD Card access

spiMaskDO               res    1
spiMaskClk              res    1
spiMaskDI               res    1
spiMaskInc              res    1

dira_backup             res     1               ' RV: dira backup 
outa_backup             res     1               ' RV: outa backup 

' From Winbond:
flashTemp               res     1
flashMask               res     1
flashpreamble           res     4                       ' Spin preamble
readStride              res     1
readStrideCount         res     1
readSkip                res     1
readDirection           res     1

                        fit     $1F0